#!/usr/bin/env python

"""
A simple simulation of an Enigma machine

Most of the complexity is in the rotor class.
The plugboard is a simple bi-directional transformation.

By default we emulate a standard 3 rotor machine,
but can also emulate a 4 or more rotor machine.

By using the 'thin B' reflector and 'Beta' rotor in the 'A' position,
we can emulate a 4 rotor machine emulating a 3 rotor machine ;-)

See http://www.crufty.net/sjg/blog/EnigmaEmulator.htm
"""

from __future__ import print_function

"""
RCSid:
	$Id: enigma.py,v 1.23 2021/03/04 06:41:05 sjg Exp $

	@(#) Copyright (c) 2012 Simon J. Gerraty

	This file is provided in the hope that it will
	be of use.  There is absolutely NO WARRANTY.
	Permission to copy, redistribute or otherwise
	use this file is hereby granted provided that 
	the above copyright notice and this notice are
	left intact. 
      
	Please send copies of changes and bug-fixes to:
	sjg@crufty.net

"""

import sys
from rotor import *
from plugboard import *


class enigma:
    def __init__(self, rotors=[], plugs=[], conf={}):
        """create an Enigma machine using the supplied rotors and plug pairs"""
        self.debug = int(conf.get('debug', 0))
        self.group = int(conf.get('group', 5))
        self.group_nl = 10 * self.group
        self.rotors = rotors
        self.pb = plugboard(plugs)
        for i in range(len(rotors)):
            if i > 0:
                rotors[i].next = rotors[i-1]
                rotors[i-1].previous = rotors[i]
        # we only interact with the right most rotor
        self.rr = rotors[-1]
        self.key = None
        self.count = 0

    def set_key(self, key):
        """Set each rotor to its initial position"""
        self.key = key
        # we skip 0th rotor[0] which is the reflector
        for i in range(len(key)):
            self.rotors[i+1].set_position(key[i].upper())

    def process(self, c):
        """Each input char is first transformed by the pluggboard,
        then the right most rotor (after stepping), which in turn invokes
        the other rotors forwards and backwards.
        Finally the plugboard has a final chance to transform the result
        from the rotors."""
        if not c.isalpha():
            return c
        c = c.upper()
        r = self.pb.transform(c)
        if self.debug and c != r:
            print('In: plugged: {0} -> {1}'.format(c, r))
        if self.debug:
            print()
        self.rr.rotate()
        r = self.rr.transform(r)
        c = self.pb.transform(r)
        if self.debug:
            if c != r:
                print('Out: plugged: {0} -> {1}'.format(r, c))
            print(self.rr.display())
        return c

    def cipher(self, str):
        """process a string of input and output in 4 letter groups"""
        t = []
        for i in range(len(str)):
            if str[i].isspace():
                continue
            r = self.process(str[i])
            if self.debug:
                print('{0} -> {1}'.format(str[i], r))
            t.append(r)
            self.count += 1
            if self.count % self.group_nl == 0:
                t.append('\n')
            elif self.count % self.group == 0:
                t.append(' ')
        return ''.join(t)
    
if __name__ == '__main__':
    import getopt

    nrotors = 0
    key = None
    Olist = None
    Rlist = None
    Reflector = None
    pairs = []
    debug = 0
    A = ord('A')
    conf = {}

    opts, args = getopt.getopt(sys.argv[1:], 'dg:O:R:r:K:P:S:n:')
    for o, a in opts:
        if o == '-d':
            debug += 1
        elif o == '-g':
            conf['group'] = int(a)
        elif o == '-K':
            key = a
        elif o == '-R':
            Rlist = a.split(',')
        elif o == '-r':
            Reflector = a
        elif o == '-O':
            Olist = a.split(',')
        elif o == '-P':
            if a.find(' ') > 0:
                Plist = a.split(' ')
            else:
                Plist = a.split(',')
            for pair in Plist:
                pairs.append(pair)
        elif o == '-n':
            nrotors = int(a)
        elif o == '-S':
            # another idea from Mike Koss's emulator
            # but we handle more rotors.
            import random

            if not nrotors:
                nrotors = 3
            random.seed(a)

            Rlist = random.sample(['I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII'], nrotors)
            Olist = random.sample(range(26), nrotors)
            # use a random number of plug pairs
            npairs = random.randint(0, 13) * 2
            if npairs > 0:
                plist = random.sample(range(26), npairs)
                for i in range(0,npairs,2):
                    pairs.append(chr(A+plist[i])+chr(A+plist[i+1]))
            key = ''
            for i in random.sample(range(26), nrotors):
                key += chr(A + i)

            print(Rlist)
            print(Olist)
            print(pairs)
            print(key)


    if not Rlist:
        Rlist = ['I', 'II', 'III']

    nrotors = len(Rlist)

    conf['debug'] = debug

    if not key:
        key = 'A' * nrotors

    if not Olist:
        Olist = [0] * nrotors

    try:
        x = int(Olist[0])
        for i in range(len(Olist)):
            Olist[i] = int(Olist[i])

    except:
        for i in range(len(Olist)):
            Olist[i] = ord(Olist[i]) - A

    if nrotors > 3:
        if not Reflector:
            Reflector = 'thin B'

    if not Reflector:
        Reflector = 'B'
        
    rotors = []
    rotors.append(rotor('R', Rotors[Reflector], debug=debug))
    for i in range(nrotors):
        rotors.append(rotor('{0}'.format(nrotors - i), Rotors[Rlist[i]], Olist[i], debug=debug))

    if debug:
        print('Type     | A B C D E F G H I J K L M N O P Q R S T U V W X Y Z | Turnover')
        for r in rotors:
            print(r)
    E = enigma(rotors, pairs, conf)
    E.set_key(key)

    for w in args:
        if w == '-':
            for line in sys.stdin:
                sys.stdout.write(E.cipher(line.rstrip()))
        else:
            sys.stdout.write(E.cipher(w))
    print()
