"""
A simple simulation of an Enigma rotor

The rotor transforms in the Rotors dictionary are from
http://en.wikipedia.org/wiki/Enigma_rotor_details

"""

from __future__ import print_function

"""
RCSid:
	$Id: rotor.py,v 1.23 2021/03/04 06:41:05 sjg Exp $

	@(#) Copyright (c) 2012 Simon J. Gerraty

	This file is provided in the hope that it will
	be of use.  There is absolutely NO WARRANTY.
	Permission to copy, redistribute or otherwise
	use this file is hereby granted provided that 
	the above copyright notice and this notice are
	left intact. 
      
	Please send copies of changes and bug-fixes to:
	sjg@crufty.net

"""

import sys

# Rotor transforms and turnover positions per wikipedia
Rotors = {
    'I': { 'type': 'I', 'transform': 'EKMFLGDQVZNTOWYHXUSPAIBRCJ', 'turnover': 'Q'},
    'II': { 'type': 'II', 'transform': 'AJDKSIRUXBLHWTMCQGZNPYFVOE', 'turnover': 'E'},
    'III': { 'type': 'III', 'transform': 'BDFHJLCPRTXVZNYEIWGAKMUSQO', 'turnover': 'V'},
    'IV': { 'type': 'IV', 'transform': 'ESOVPZJAYQUIRHXLNFTGKDCMWB', 'turnover': 'J'},
    'V': { 'type': 'V', 'transform': 'VZBRGITYUPSDNHLXAWMJQOFECK', 'turnover': 'Z'},
    'VI': { 'type': 'VI', 'transform': 'JPGVOUMFYQBENHZRDKASXLICTW', 'turnover': 'ZM'},
    'VII': { 'type': 'VII', 'transform': 'NZJHGRCXMYSWBOUFAIVLPEKQDT', 'turnover': 'ZM'},
    'VIII': { 'type': 'VIII', 'transform': 'FKQHTLXOCBJSPDZRAMEWNIUYGV', 'turnover': 'ZM'},
    'A': { 'type': 'A', 'transform': 'EJMZALYXVBWFCRQUONTSPIKHGD', 'turnover': None},
    'B': { 'type': 'B', 'transform': 'YRUHQSLDPXNGOKMIEBFZCWVJAT', 'turnover': None},
    'C': { 'type': 'C', 'transform': 'FVPJIAOYEDRZXWGCTKUQSBNMHL', 'turnover': None},
    'thin B': { 'type': 'thin B', 'transform': 'ENKQAUYWJICOPBLMDXZVFTHRGS', 'turnover': None},
    'thin C': { 'type': 'thin C', 'transform': 'RDOBJNTKVEHMLFCWZAXGYIPSUQ', 'turnover': None},
    'Beta': { 'type': 'Beta', 'transform': 'LEYJVCNIXWPBQMDRTAKZGFUHOS', 'turnover': None},
    'Gamma': { 'type': 'Gamma', 'transform': 'FSOKANUERHMBTIYCWLQPZXVGJD', 'turnover': None},
    'ETW': { 'type': 'ETW', 'transform': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', 'turnover': None},
    'IC': { 'type': 'IC', 'transform': 'DMTWSILRUYQNKFEJCAZBPGXOHV',  'turnover': None},
    'IIC': { 'type': 'IIC', 'transform': 'HQZGPJTMOBLNCIFDYAWVEUSRKX',  'turnover': None},
    'IIIC': { 'type': 'IIIC', 'transform': 'UQNTLSZFMREHDPXKIBVYGJCWOA',  'turnover': None},
    #'': { 'type': '', 'transform': '',  'turnover': None},
    }

class rotor:
    A = ord('A')
    ring = []
    
    def __init__(self, name, conf={}, offset=0, previous=None, next=None, debug=0):
        """create rotor instance using transform and turnover points from conf

        Many rotor based cipher machines were in use at various times.
        We want to be able to emulate any of them.
        Of course that requires knowing the rotor details.
        """
        
        self.type = conf.get('type')
        transform = conf.get('transform')
        turnover  = conf.get('turnover')
        self.name = name           # 1==Right most R=reflector
        self.debug = debug
        if name[0] == 'R':         # Reflector
            self.number = -1
        else:
            self.number = int(name)# rotors 1(Right),2(middle),3(left)...
        # for enigma, reflector and rotor 4 do not rotate
        if self.number < 0 or self.number > 3:
            rotates = False
        else:
            rotates = True
        self.rotates = conf.get('rotates', rotates)
        self.trans = []            # the transformation
        self.fwd = {}              # forward map
        self.rev = {}              # reverse map
        self.offset = offset       # offset of transformation from alpha ring
        self.mod = len(transform)  # length of trans
        self.turnover = []         # positions at which we rotate next
        self.previous = previous   # previous rotor if any
        self.next = next           # next rotor if any
        self.position = -1         # not initialized yet

        if len(self.ring) == 0:
            a = self.A
            for i in range(self.mod):
                self.ring.append(chr(a+i))

        for i in range(self.mod):
            c = transform[i]
            # this is just for display
            self.trans.append(c)
            x = ord(c) - self.A
            # These maps provide the conversion factors.
            # Rather than letter substitutions, these contain
            # ordinal shifts eg. A->E is (0+4)%26 and E->A is (26-4)%26
            # keeping in mind that the value will be added to the ordinal
            # value of the incomming letter and again mod 26, so
            # an incomming E is (4+22)%26 == 0 which is 'A'. 
            # Unlike letter substitutions, this model is unaffected by
            # ring offsets.
            self.fwd[i] = (x - i) % self.mod
            self.rev[x] = (i - x) % self.mod
            if debug > 2:
                print('{6}: i={0} a={1} c={2} x={3} fwd={4} rev={5}'.format(i,self.ring[i],c,x,self.fwd[i],self.rev[x],self.type))

        if self.number < 3 and turnover:
            for i in range(len(turnover)):
                x = self.ring.index(turnover[i])
                self.turnover.append(x)

            if debug > 1:
                print('{0}: turnover: {1}'.format(self.name, self.turnover))

        if debug > 1:
            print(self)
            if debug > 2:
                print(self.fwd)
                print(self.rev)
            
    def set_position(self, position):
        """initialize the position of a rotor
        This is how the session key is set."""
        self.position = self.ring.index(position)
        if self.debug:
            print('{0}: position {1} {2}'.format(self.name, position, self.position))


    def transform(self, c, backwards=False):
        """Transform c according to this rotor's wiring.
        Pass on to previous or next rotor based on value of backwards.
        If we are the reflector we set backwards True."""

        if self.position < 0:
            self.set_position('A')
        i = ord(c) - self.A
        if backwards:
            tr = self.rev
            nr = self.previous
        else:
            tr = self.fwd
            nr = self.next
            if self.number < 0:
                backwards = True
                nr = self.previous

        x = tr[(i + self.position - self.offset) % self.mod]
        r = self.ring[(i + x) % self.mod]

        if self.debug:
            if self.debug > 1:
                print('{6}: {0} -> {1} (i = {2:2}, shift = {3:2}, position = {4:2}, offset = {5:2})'.format(c,r,i,x,self.position,self.offset,self.name))
            else:
                print('{2}: {0} -> {1}'.format(c,r,self.name))

        if nr:
            r = nr.transform(r, backwards)
        return r

    def display(self):
        """indicate the letter that would be visible for this rotor.
        The request is passed on to the next rotor first."""
        if self.number < 0:             # reflector
            return ''
        if self.next:
            r = self.next.display()
        else:
            r = ''
        return r + self.ring[self.position]

    def rotate(self):
        """rotate this rotor one step.
        Check the turnover notches (if any) and call next rotors
        rotate method if needed.
        Only the 3 right most rotors step.
        The 2nd rotor can double step."""

        if not self.rotates:
            return

        # we may need to double-step rotor 2
        if self.next:
            if self.position in self.turnover or (self.number == 1 and self.next.position in self.next.turnover):
                self.next.rotate()
        self.position += 1
        self.position %= self.mod

        if self.debug:
            print('{0}: rotate -> {1} [{2}]'.format(self.name,
                                                    self.position,
                                                    self.ring[self.position]))

    def __str__(self):
        return '{0:<8} | {1} | {2} ({3})'.format(self.type,
                                                 ' '.join(self.trans),
                                                 self.turnover, self.offset)

if __name__ == '__main__':
    import getopt

    def rep_rotor(name, transform, turnover):
        t=[]
        for i in range(len(transform)):
            t.append(transform[i])
        return '{0:<8} | {1} | {2}'.format(name, ' '.join(t), turnover)

    def list_rotors(kl):
        print(rep_rotor('Type', 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', 'Turnover'))
        for k in kl:
            r = Rotors[k]
            print(rep_rotor(r['type'], r['transform'], r['turnover']))

    R = 'I'
    K = 'A'
    Oset = 0
    backwards = False
    debug = 0
    
    opts, args = getopt.getopt(sys.argv[1:], 'dO:K:R:rLl')
    for o, a in opts:
        if o == '-d':
            debug=1
        elif o == '-K':
            K = a
        elif o == '-R':
            R = a
        elif o == '-r':
            backwards = True
        elif o == '-O':
            Oset = int(a)
        elif o == '-L':
            list_rotors(sorted(Rotors.keys()))
        elif o == '-l':
            list_rotors(['I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII', 'B'])

    r0 = rotor('1', Rotors[R], Oset, debug=debug)
    r0.set_position(K)
    if debug:
        print(rep_rotor('Type', 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', 'Turnover'))

        print(r0)
        print(r0.fwd)
        print(r0.rev)
    for w in args:
        for i in range(len(w)):
            r = r0.transform(w[i], backwards)
            print('{0} -> {1}'.format(w[i], r))
